@@ -82,11 +82,10 @@ static shell::logmode_t errmode = shell::NONE;
 static const char *errname = NULL;
 static shell::logproc_t errproc = (shell::logproc_t)NULL;
 static mutex_t symlock;
-static shell::downproc_t downproc = (shell::downproc_t)NULL;
 static char **_orig = NULL;
-
-OrderedIndex shell::Option::index;
-const char *shell::domain = NULL;
+static OrderedIndex _index;
+static const char *_domain = NULL;
+static shell::exitproc_t _exitproc = NULL;
 
 shell::pipeio::pipeio()
 {
@@ -395,7 +394,7 @@ void shell::iobuf::open(const char *path, char **argv, shell::pmode_t mode, size
 }
 
 shell::Option::Option(char shortopt, const char *longopt, const char *value, const char *help) :
-OrderedObject(&index)
+OrderedObject(&_index)
 {
     while(longopt && *longopt == '-')
         ++longopt;
@@ -411,6 +410,16 @@ shell::Option::~Option()
 {
 }
 
+void shell::Option::reset(void)
+{
+    _index.reset();
+}
+
+LinkedObject *shell::Option::first(void)
+{
+    return _index.begin();
+}
+
 void shell::Option::disable(void)
 {
     short_option = 0;
@@ -585,7 +594,7 @@ void shell::set0(char *argv0)
         *ext = 0;
 #endif
 
-    if(!domain)
+    if(!_domain)
         bind(_argv0);
 }
 
@@ -1214,8 +1223,8 @@ const char *shell::getenv(const char *id, const char *value)
 
     if(errname)
         keyid = errname;
-    else if(domain)
-        keyid = domain;
+    else if(_domain)
+        keyid = _domain;
     else if(_argv0)
         keyid = _argv0;
 
@@ -1379,64 +1388,55 @@ exit:
     return pid;
 }
 
-static SERVICE_STATUS_HANDLE hStatus;
-static SERVICE_STATUS status;
 static unsigned detaches = 0;
-
-static void WINAPI control(DWORD control)
-{
-    switch(control) {
-    case 128:
-        if(downproc)
-            (*downproc)(true);
-        return;
-    case SERVICE_CONTROL_SHUTDOWN:
-    case SERVICE_CONTROL_STOP:
-        status.dwCurrentState = SERVICE_STOP_PENDING;
-        status.dwWin32ExitCode = 0;
-        status.dwCheckPoint = 0;
-        status.dwWaitHint = 6000;
-        SetServiceStatus(hStatus, &status);
-        if(downproc)
-            (*downproc)(false);
-        else
-            exit(0);
-        break;
-    default:
-        break;
-    }
-}
-
-extern "C" {
-    extern int main(int argc, char **argv);
-}
+static shell::mainproc_t _entry = NULL;
 
 static void WINAPI _start(DWORD argc, LPSTR *argv)
 {
     ++detaches;
-    ::main(argc, argv);
+    (*_entry)(argc, argv);
 }
 
-void shell::detach(downproc_t stop)
+static void exit_handler(void)
+{
+    if(_exitproc) {
+        (*_exitproc)();
+        _exitproc = NULL;
+    }
+}
+
+static BOOL WINAPI _stop(DWORD code)
+{
+    exit_handler();
+    return true;
+}
+
+void shell::exiting(exitproc_t handler)
+{
+    if(!_exitproc && handler) {
+        _exitproc = handler;
+        if(_domain)
+            SetConsoleTitle(_domain);
+        SetConsoleCtrlHandler((PHANDLER_ROUTINE)_stop, TRUE);
+        atexit(exit_handler);
+    }
+    else
+        _exitproc = handler;
+}
+
+void shell::detach(mainproc_t entry)
 {
     const char *name = _argv0;
 
-    downproc = stop;
-
-    if(domain)
-        name = domain;
+    if(_domain)
+        name = _domain;
 
     name = strdup(name);
 
-    if(detaches) {
-        memset(&status, 0, sizeof(SERVICE_STATUS));
-        status.dwServiceType = SERVICE_WIN32;
-        status.dwCurrentState = SERVICE_START_PENDING;
-        status.dwControlsAccepted = SERVICE_ACCEPT_STOP|SERVICE_ACCEPT_SHUTDOWN;
-
-        hStatus = ::RegisterServiceCtrlHandler(name, &control);
+    if(detaches || entry == NULL)
         return;
-    }
+
+    _entry = entry;
 
     SERVICE_TABLE_ENTRY servicetable[] = {
         {(LPSTR)name, &_start},
@@ -1451,14 +1451,6 @@ void shell::restart(void)
 {
 }
 
-void shell::up(void)
-{
-    if(detaches) {
-        status.dwCurrentState = SERVICE_RUNNING;
-        ::SetServiceStatus(hStatus, &status);
-    }
-}
-
 int shell::detach(const char *path, char **argv, char **envp, fd_t *stdio)
 {
     STARTUPINFO si;
@@ -1614,18 +1606,17 @@ int shell::system(const char *cmd, const char **envp)
         ++envp;
     }
 
+    ::signal(SIGHUP, SIG_DFL);
+    ::signal(SIGABRT, SIG_DFL);
     ::signal(SIGQUIT, SIG_DFL);
     ::signal(SIGINT, SIG_DFL);
     ::signal(SIGCHLD, SIG_DFL);
     ::signal(SIGPIPE, SIG_DFL);
+    ::signal(SIGUSR1, SIG_DFL);
     ::execlp("/bin/sh", "sh", "-c", cmd, NULL);
     ::exit(-1);
 }
 
-void shell::up(void)
-{
-}
-
 void shell::restart(void)
 {
     pid_t pid;
@@ -1654,14 +1645,39 @@ restart:
     }
 }
 
-void shell::detach(downproc_t stop)
+static void exit_handler(void)
+{
+    if(_exitproc) {
+        (*_exitproc)();
+        _exitproc = NULL;
+    }
+}
+
+static void abort_handler(int signo)
+{
+    exit_handler();
+}
+
+void shell::exiting(exitproc_t handler)
+{
+
+    if(!_exitproc && handler) {
+        _exitproc = handler;
+        ::signal(SIGABRT, abort_handler);
+#ifdef  HAVE_ATEXIT
+        ::atexit(exit_handler);
+#endif
+    }
+    else
+        _exitproc = handler;
+}
+
+void shell::detach(mainproc_t entry)
 {
     const char *dev = "/dev/null";
     pid_t pid;
     int fd;
 
-    downproc = stop;
-
     close(0);
     close(1);
     close(2);
@@ -1739,7 +1755,9 @@ int shell::detach(const char *path, char **argv, char **envp, fd_t *stdio)
     ::signal(SIGINT, SIG_DFL);
     ::signal(SIGCHLD, SIG_DFL);
     ::signal(SIGPIPE, SIG_DFL);
-    ::signal(SIGHUP, SIG_IGN);
+    ::signal(SIGHUP, SIG_DFL);
+    ::signal(SIGABRT, SIG_DFL);
+    ::signal(SIGUSR1, SIG_DFL);
 
 #ifdef  SIGTTOU
     ::signal(SIGTTOU, SIG_IGN);
@@ -1839,6 +1857,9 @@ shell::pid_t shell::spawn(const char *path, char **argv, char **envp, fd_t *stdi
     ::signal(SIGINT, SIG_DFL);
     ::signal(SIGCHLD, SIG_DFL);
     ::signal(SIGPIPE, SIG_DFL);
+    ::signal(SIGHUP, SIG_DFL);
+    ::signal(SIGABRT, SIG_DFL);
+    ::signal(SIGUSR1, SIG_DFL);
 
     for(fd = 0; fd < 3; ++fd) {
         if(stdio && stdio[fd] != INVALID_HANDLE_VALUE)
@@ -1926,14 +1947,14 @@ void shell::bind(const char *name)
     const char *locale = UCOMMON_LOCALE;
 #endif
 
-    if(!domain) {
+    if(!_domain) {
         setlocale(LC_ALL, "");
         bindtextdomain("ucommon", locale);
     }
 
     bindtextdomain(name, locale);
     textdomain(name);
-    domain = name;
+    _domain = name;
 }
 
 void shell::rebind(const char *name)
@@ -1941,7 +1962,7 @@ void shell::rebind(const char *name)
     if(name)
         textdomain(name);
     else
-        textdomain(domain);
+        textdomain(_domain);
 }
 
 #ifdef  _MSWINDOWS_
