@@ -63,8 +63,6 @@ private:
     char *_argv0;
     LinkedObject *_syms;
 
-    static const char *domain;
-
     class __LOCAL args : public OrderedObject
     {
     public:
@@ -112,9 +110,14 @@ public:
     typedef bool (*logproc_t)(loglevel_t level, const char *text);
 
     /**
-     * Shutdown handler.
+     * Main handler.
      */
-    typedef void (*downproc_t)(bool reload);
+    typedef int (*mainproc_t)(int argc, char **argv);
+
+    /**
+     * Exit handler.
+     */
+    typedef void (*exitproc_t)(void);
 
 #ifdef  _MSWINDOWS_
     typedef HANDLE pid_t;
@@ -313,9 +316,6 @@ public:
      */
     class __EXPORT Option : public OrderedObject
     {
-    private:
-        static OrderedIndex index;
-
     public:
         char short_option;
         const char *long_option;
@@ -334,8 +334,7 @@ public:
 
         virtual ~Option();
 
-        inline static LinkedObject *first(void)
-            {return index.begin();};
+        static LinkedObject *first(void);
 
         /**
          * Disable a option.  Might happen if argv0 name suggests an
@@ -350,8 +349,7 @@ public:
          */
         virtual const char *assign(const char *value) = 0;
 
-        inline static void reset(void)
-            {index.reset();};
+        static void reset(void);
     };
 
     /**
@@ -687,11 +685,6 @@ public:
      */
     void restart(char *argv0, char **argv, char **list);
 
-    /**
-     * Notify service is up.
-     */
-    void up(void);
-
     /**
      * Parse shell arguments directly into a shell object.
      * @param args table.
@@ -801,10 +794,12 @@ public:
     inline const char *operator[](unsigned offset)
         {return _argv[offset];};
 
+    static void exiting(exitproc_t);
+
     /**
-     * Detach current process to daemon.
+     * Detach current process to daemon for service entry.
      */
-    void detach(downproc_t stopreload = (downproc_t)NULL);
+    void detach(mainproc_t mainentry = (mainproc_t)NULL);
 
     /**
      * Make current process restartable.
